#include <windows.h>

#include <initguid.h>
#include "f1addin.h"

//////////////////////////////////////////////////////////////////////////////
// These strings provide the name and description of the add-in.  The L beside
// the left quote causes the compiler to generate a Unicode string (i.e., it
// is a short array instead of a char array).
//
#ifdef _DEBUG
#define F1AddInName L"Demo Add-In (D)"
#define F1AddInDesc L"C++ Add-In Demo\n(Debug Build)"
#else
#define F1AddInName L"Demo Add-In"
#define F1AddInDesc L"C++ Add-In Demo"
#endif

//////////////////////////////////////////////////////////////////////////////
// HELPER FUNCTIONS to assist with add-in function return values.
//
// There are two categories of errors that may occur in an add-in function.
//
// The most common is an error in evaluation due to bad data or some other
// forseeable situation that can be cleanly dealt with by a simple error value
// such as #VALUE!.  The MakeErrorResult helper function returns this kind of
// error.
//
// The second is more catostrophic in nature, halting the recalc operation and
// possibly the entire application.  This normally occurs only when the system
// runs low on memory.  The CheckReturnValue function accepts any HRESULT and
// converts it into a value expected by Formula One.  Formula One 6.1 always
// expects either S_OK or E_OUTOFMEMORY.
// 
// See the sample add-in functions for examples of how these helper functions
// are used.
//
HRESULT MakeErrorResult(LPVARIANTARG pResult, HRESULT hrEval = S_OK)
{
	if (hrEval)
	{
		if (SUCCEEDED(::VariantClear(pResult)))
		{
			pResult->vt = VT_ERROR;
			pResult->scode = hrEval;
		}

		else
			return E_OUTOFMEMORY;
	}

	return S_OK;
}

inline HRESULT CheckReturnValue(HRESULT hr)
{
	return FAILED(hr) ? E_OUTOFMEMORY : S_OK;
}

//////////////////////////////////////////////////////////////////////////////
// RETURN.ME function (::ReturnMe)
//
// This function, which is the implementation of the RETURN.ME add-in
// function, is exposed to Formula One by the exported function F1AddInInit,
// which is found at the bottom of this file.
//
// RETURN.ME accepts a single argument and returns that argument's value.  If
// the argument is a range reference, the top-left cell's value is returned.
//
HRESULT CALLBACK ReturnMe(

		LPVARIANTARG pResult,	// Pointer to the VARIANT receiving the
								// function's result

		int,					// Ignore this argument; it's reserved for
								// future use

		int nArgs,				// Number of arguments

		LPVARIANTARG pArgs)		// Pointer to an array of nArgs VARIANTs, each
								// corresponding to an argument being passed
								// to the add-in function.  pArgs[0] is the
								// first (left-most) argument.
{
	if (nArgs == 1)
	{
		// VT_UNKNOWN requires special handling
		if (pArgs[0].vt == VT_UNKNOWN)
		{
			// We handle arrays; we don't know about any other interfaces, so
			// all others are ignored here.
			IF1AddInArrayEx* pArray;
			if (SUCCEEDED(pArgs[0].punkVal->QueryInterface(IID_IF1AddInArrayEx, (LPVOID*)&pArray)))
			{
				// Copy the value of the top-left cell of the first sheet or
				// range to the result.
				pArray->GetValue(0, 0, 0, pResult);
				pArray->Release();

				return S_OK;
			}
		}

		else
			// Copy the argument's value to the result.
			return ::CheckReturnValue(::VariantCopy(pResult, pArgs));
	}

	// The above code should have already handled all situations except bad
	// arguments.  The standard result for this situation is #VALUE!.
	return ::MakeErrorResult(pResult, F1_E_VALUE);
}

//////////////////////////////////////////////////////////////////////////////
// GET.IN.RANGE function (::GetInRange)
//
// This function, which is the implementation of the GET.IN.RANGE add-in
// function, is exposed to Formula One by the exported function F1AddInInit,
// which is found at the bottom of this file.
//
// GET.IN.RANGE returns an item at a specified location with a range.  For
// example:
//
// Formula                                      Cell Returned
// ------------------------------------------   -------------
// =GET.IN.RANGE(A11:C13,1,1)                   A11
//
// =GET.IN.RANGE(A11:C13,2,3)                   C12
//
// =GET.IN.RANGE(Sheet1:Sheet3!A11:C13,2,1,3)   Sheet2!C11
//
// =GET.IN.RANGE((A1:C3,A11:C13),1,3,2)         B3
//
HRESULT CALLBACK GetInRange(

		LPVARIANTARG pResult,	// Pointer to the VARIANT receiving the
								// function's result

		int,					// Ignore this argument; it's reserved for
								// future use

		int nArgs,				// Number of arguments

		LPVARIANTARG pArgs)		// Pointer to an array of nArgs VARIANTs, each
								// corresponding to an argument being passed
								// to the add-in function.  pArgs[0] is the
								// first (left-most) argument.
{
	// Three arguments: a two-dimensional array
	if (nArgs == 3)
	{
		IF1AddInArray* pArray;

		// Try to get the array interface
		if (pArgs[0].vt == VT_UNKNOWN
				&& SUCCEEDED(pArgs[0].punkVal->QueryInterface(IID_IF1AddInArray, (LPVOID*)&pArray)))
		{
			// Both coordinates must be of type double
			if (pArgs[1].vt == VT_R8 && pArgs[2].vt == VT_R8)
			{
				// If the coordinates are invalid, GetValue will put #REF! in
				// pResult, which is what we want.
				pArray->GetValue((int)pArgs[1].dblVal - 1, (int)pArgs[2].dblVal - 1, pResult);
				pArray->Release();

				return S_OK;
			}

			// Release the array; we can't use it.
			pArray->Release();
		}
	}

	// Four arguments: a three-dimensional array
	else if (nArgs == 4)
	{
		IF1AddInArrayEx* pArray;

		// Try to get the array interface
		if (pArgs[0].vt == VT_UNKNOWN
				&& SUCCEEDED(pArgs[0].punkVal->QueryInterface(IID_IF1AddInArrayEx, (LPVOID*)&pArray)))
		{
			// All coordinates must be of type double
			if (pArgs[1].vt == VT_R8 && pArgs[2].vt == VT_R8 && pArgs[3].vt == VT_R8)
			{
				// If the coordinates are invalid, GetValue will put #REF! in
				// pResult, which is what we want.
				pArray->GetValue((int)pArgs[1].dblVal - 1, (int)pArgs[2].dblVal - 1, (int)pArgs[3].dblVal - 1, pResult);
				pArray->Release();

				return S_OK;
			}

			// Release the array; we can't use it.
			pArray->Release();
		}
	}

	// The above code should have already handled all situations except bad
	// arguments.  The standard result for this situation is #VALUE!.
	return ::MakeErrorResult(pResult, F1_E_VALUE);
}

//////////////////////////////////////////////////////////////////////////////
// F1AddInInit
//
// This exported function (see AddIn.def) registers the add-in's functions and
// other general information with Formula One.  Formula One calls this
// function when it loads the add-in.
//
HRESULT __stdcall F1AddInInit(
		F1AddInRegisterInfoProc		RegisterInfoProc,
		F1AddInRegisterFunctionProc	RegisterFunctionProc,
		int,
		int)
{
	HRESULT hr = S_OK;

	// Register the RETURN.ME function
	if (SUCCEEDED(hr))
		hr = RegisterFunctionProc(
				L"RETURN.ME",// Name of the function as seen by the end-user
				0,			// Reserved for future use; must be 0
				ReturnMe,	// Name of the function as implemented in the DLL
				1);			// Number of arguments expected.

	// Register the GET.IN.RANGE function
	if (SUCCEEDED(hr))
		hr = RegisterFunctionProc(
				L"GET.IN.RANGE",// Name of the function as seen by the end-user
				0,			// Reserved for future use; must be 0
				GetInRange,	// Name of the function as implemented in the DLL
				-1);		// -1 because this function may accept a variable
							//     number of arguments

	// Register general add-in information
	if (SUCCEEDED(hr))
		hr = RegisterInfoProc(
				F1AddInName,// Name of the add-in (defined at top of the file)
				F1AddInDesc,// Description of the add-in
				0,			// Reserved for future use; must be 0
				0);			// Reserved for future use; must be 0

	return hr;
}

//////////////////////////////////////////////////////////////////////////////
// DllMain
//
// Initialization may be done here or in F1AddInInit--whichever best suits
// your needs.  See the Win32 API doc for more details about this function.
//
BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}
